/*
    OCLTools is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.

    Dr. Zaius
    ClusterChimps.org
*/

// Multiply two matrices A * B = C

#include <stdlib.h>
#include <stdio.h>
#include <math.h>
#include <oclutil.h>
#include <CL/cl.h>

#define WA 3
#define HA 3
#define WB 3
#define HB 3
#define WC 3
#define HC 3

// Allocates a matrix with random float entries.
void randomInit(float* data, int size)
{
   for (int i = 0; i < size; ++i)
   data[i] = rand() / (float)RAND_MAX;
}



/////////////////////////////////////////////////////////
// Program main
/////////////////////////////////////////////////////////

int
main(int, char**)
{

   // set seed for rand()
   srand(2006);

   // 1. allocate host memory for matrices A and B
   int size_A = WA * HA;
   int mem_size_A = sizeof(float) * size_A;
   float* h_A = (float*) malloc(mem_size_A);

   int size_B = WB * HB;
   int mem_size_B = sizeof(float) * size_B;
   float* h_B = (float*) malloc(mem_size_B);

   // 2. initialize host memory
   randomInit(h_A, size_A);
   randomInit(h_B, size_B);

   // 3. print out A and B
   printf("\n\nMatrix A\n");
   for(int i = 0; i < size_A; i++)
   {
      printf("%f ", h_A[i]);
      if(((i + 1) % WA) == 0)
      printf("\n");
   }

    printf("\n\nMatrix B\n");
   for(int i = 0; i < size_B; i++)
   {
      printf("%f ", h_B[i]);
      if(((i + 1) % WB) == 0)
      printf("\n");
   }

   // 4. allocate host memory for the result C
   int size_C = WC * HC;
   int mem_size_C = sizeof(float) * size_C;
   float* h_C = (float*) malloc(mem_size_C);

   // 5. Initialize OpenCL
   // OpenCL specific variables
   cl_context       clGPUContext;
   cl_command_queue clCommandQue;
   cl_program       clProgram;
   cl_kernel        clKernel;
   cl_device_id     cdDevice;

   size_t kernelLength;
   cl_int errcode;

   // OpenCL device memory for matrices
   cl_mem d_A;
   cl_mem d_B;
   cl_mem d_C;

   cl_platform_id   cpPlatform;

   errcode = ocltGetPlatformID(&cpPlatform, "NVIDIA");
   checkError(errcode, CL_SUCCESS);

   // Get a GPU device
   errcode = clGetDeviceIDs(cpPlatform, CL_DEVICE_TYPE_GPU, 1, &cdDevice, NULL);
   checkError(errcode, CL_SUCCESS);

   clGPUContext = clCreateContext(0, 1, &cdDevice, NULL, NULL, &errcode);
   checkError(errcode, CL_SUCCESS);

   //Create a command-queue
   clCommandQue = clCreateCommandQueue(clGPUContext, cdDevice, 0,
                                       &errcode);
   checkError(errcode, CL_SUCCESS);

   // Setup device memory
   d_C = clCreateBuffer(clGPUContext, CL_MEM_READ_WRITE, mem_size_A, NULL,
                        &errcode);
   d_A = clCreateBuffer(clGPUContext, CL_MEM_READ_WRITE |
                        CL_MEM_COPY_HOST_PTR, mem_size_A, h_A, &errcode);
   d_B = clCreateBuffer(clGPUContext, CL_MEM_READ_WRITE |
                        CL_MEM_COPY_HOST_PTR, mem_size_B, h_B, &errcode);

   // 6. Load and build OpenCL kernel
   ocltExtractKernels();
   
    size_t lSize;
   char *compilerFlags;
   unsigned char *buffer = ocltGetEmbeddedKernelBin("matrixMul", &compilerFlags, &lSize);

   cl_int status;
   clProgram = clCreateProgramWithBinary(clGPUContext, 1,
                 (const cl_device_id *)&cdDevice, &lSize,
                 (const unsigned char**)&buffer, &status, &errcode);

   errcode = clBuildProgram(clProgram, 0, NULL, compilerFlags, NULL, NULL);
   checkError(errcode, CL_SUCCESS);

   clKernel = clCreateKernel(clProgram, "matrixMul", &errcode);
   checkError(errcode, CL_SUCCESS);


   // 7. Launch OpenCL kernel
   size_t localWorkSize[2], globalWorkSize[2];

   int wA = WA;
   int wC = WC;
   errcode  = clSetKernelArg(clKernel, 0, sizeof(cl_mem), (void *)&d_C);
   errcode |= clSetKernelArg(clKernel, 1, sizeof(cl_mem), (void *)&d_A);
   errcode |= clSetKernelArg(clKernel, 2, sizeof(cl_mem), (void *)&d_B);
   errcode |= clSetKernelArg(clKernel, 3, sizeof(int), (void *)&wA);
   errcode |= clSetKernelArg(clKernel, 4, sizeof(int), (void *)&wC);
   checkError(errcode, CL_SUCCESS);

   localWorkSize[0] = 3;
   localWorkSize[1] = 3;
   globalWorkSize[0] = 3;
   globalWorkSize[1] = 3;

   errcode = clEnqueueNDRangeKernel(clCommandQue, clKernel, 2, NULL,
                                    globalWorkSize, localWorkSize, 0, NULL,
                                    NULL);
   checkError(errcode, CL_SUCCESS);

   // 8. Retrieve result from device
   errcode = clEnqueueReadBuffer(clCommandQue, d_C, CL_TRUE, 0, mem_size_C,
                                 h_C, 0, NULL, NULL);
   checkError(errcode, CL_SUCCESS);

   // 9. print out the results
   printf("\n\nMatrix C (Results)\n");
   for(int i = 0; i < size_C; i++)
   {
      printf("%f ", h_C[i]);
      if(((i + 1) % WC) == 0)
      printf("\n");
   }
   printf("\n");

   // 10. clean up memory
   free(h_A);
   free(h_B);
   free(h_C);

   clReleaseMemObject(d_A);
   clReleaseMemObject(d_C);
   clReleaseMemObject(d_B);

   free(buffer);
   clReleaseContext(clGPUContext);
   clReleaseKernel(clKernel);
   clReleaseProgram(clProgram);
   clReleaseCommandQueue(clCommandQue);

}

